local name, private = ...

local list = private.objects.list
local dict = private.objects.dict

private.objects.ui = {}
private.objects.ui.base = {}

local L = private.L

------------------------------------------------------------
-------------------------- Label ---------------------------
------------------------------------------------------------

private.objects.ui.label = private.objects.base()

function private.objects.ui.label:init(parent, text, size)
    self.frame = parent:CreateFontString()
    self.frame:SetFont("Fonts\\FRIZQT__.TTF", size or 12)
    self.frame:SetText(text)
end

function private.objects.ui.label:SetText(text)
    self.frame:SetText(text)
end

------------------------------------------------------------
--------------------- Item link label ----------------------
------------------------------------------------------------

private.objects.ui.itemLinkLabel = private.objects.base()

function private.objects.ui.itemLinkLabel:init(parent, itemID)
    self.frame = CreateFrame("Button", nil, parent)
    self.frame:SetNormalFontObject('GameFontHighlightLeft')
    self.frame.__parent = self
    self.frame:SetSize(250, 15)
    
    self.text = ""
    self.hasItem = true
    
    self.item = {}
    self.item.ID = itemID
    self.item.name, self.item.link = GetItemInfo(itemID)
    
    self.frame:SetText(self.item.link)
    
    self.frame:RegisterForClicks("LeftButtonUp", "RightButtonUp")
    self.frame:SetScript("OnEnter", self.ShowTooltip)
    self.frame:SetScript("OnLeave", self.HideTooltip)
    self.frame:SetScript("OnClick", self.frameclick)
end

function private.objects.ui.itemLinkLabel:SetItem(itemID)
    -- Sets the item for the label to a different item
    
    -- Check if the "itemID" is numeric - if not, it's text, and treat it as such
    if tonumber(itemID) then
        self.item.ID = itemID
        self.item.name, self.item.link = GetItemInfo(itemID)
        self.frame:SetText(self.item.link)
        self.hasItem = true
        self.frame:Enable()
    else
        self:SetText(itemID)
    end
end

function private.objects.ui.itemLinkLabel:SetText(text)
    -- Sets this label to show text instead of an item
    -- Will disable clicking and tooltips
    self.text = text
    self.frame:SetText(text)
    self.hasItem = false
    self.frame:Disable()
end

function private.objects.ui.itemLinkLabel:ShowTooltip()
    if not self.__parent.hasItem then return end
    
    GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
    GameTooltip:ClearLines()
    GameTooltip:SetHyperlink(self.__parent.item.link)
    GameTooltip:Show()
end

function private.objects.ui.itemLinkLabel:HideTooltip()
    GameTooltip:Hide()
end

function private.objects.ui.itemLinkLabel:frameclick(...)
    -- DO NOT OVERRIDE (unless you know what you're doing and need to for whatever reason)
    -- This simply pushes the click function from the button to the handler object
    self.__parent:click(...)
end

function private.objects.ui.itemLinkLabel:click(button)
    -- Reacts to any click at all
    -- Override this function if you only want one click function for both buttons
    -- Otherwise, override the individual left and right click functions
    if button == "LeftButton" then self:leftclick() end
    if button == "RightButton" then self:rightclick() end
end

function private.objects.ui.itemLinkLabel:leftclick()
    if not self.hasItem then return end
    
    if self.item.link then ChatEdit_InsertLink(self.item.link) end
end

function private.objects.ui.itemLinkLabel:rightclick()
    -- Reacts to a right click
    -- Override this to handle right clicks manually
    -- If textureButton:click() is overridden without calling this function,
    -- this will not be called on clicks.
    
    -- This is a placeholder function; override it for functionality
end

------------------------------------------------------------
----------------------- Count label ------------------------
------------------------------------------------------------

private.objects.ui.countlabel = private.objects.base()

function private.objects.ui.countlabel:init(parent, size)
    self.frame = parent:CreateFontString()
    self.frame:SetFont("Fonts\\FRIZQT__.TTF", size or 12)
    self.frame:SetText("0 / 0")
    
    self.left = 0
    self.right = 0
end

function private.objects.ui.countlabel:SetLeft(text)
    self.left = text
    self.frame:SetText(self.left .. " / " .. self.right)
end

function private.objects.ui.countlabel:SetRight(text)
    self.right = text
    self.frame:SetText(self.left .. " / " .. self.right)
end

function private.objects.ui.countlabel:SetBoth(left, right)
    self.left = left
    self.right = right
    self.frame:SetText(self.left .. " / " .. self.right)
end

------------------------------------------------------------
---------------------- Texture button ----------------------
------------------------------------------------------------

private.objects.ui.textureButton = private.objects.base()

function private.objects.ui.textureButton:init(parent, texture, name, tooltip)
    self.name = name or "Texture button"
    self.tooltip = tooltip or false
    
    self.frame = CreateFrame("Button", nil, parent or UIParent)
    self.frame.__parent = self
    self.texture = {}
    
    self.texture.normal = self.frame:CreateTexture()
    self.texture.normal:SetTexture(texture)
    self.texture.normal:SetDrawLayer("BACKGROUND")
    self.texture.normal:SetAllPoints()
    
    self.texture.pushed = self.frame:CreateTexture()
    self.texture.pushed:SetTexture("Interface/Buttons/UI-Quickslot-Depress")
    self.texture.pushed:SetAllPoints()
    self.texture.pushed:SetBlendMode("ADD")
    self.texture.pushed:SetDrawLayer("BORDER")
    self.frame:SetPushedTexture(self.texture.pushed)
    
    self.texture.disabled = self.frame:CreateTexture()
    self.texture.disabled:SetTexture(texture)
    self.texture.disabled:SetAllPoints()
    self.texture.disabled:SetDesaturated(true)
    self.frame:SetDisabledTexture(self.texture.disabled)
    
    self.texture.highlight = self.frame:CreateTexture()
    self.texture.highlight:SetTexture("Interface/Buttons/ButtonHilight-Square")
    self.texture.highlight:SetAllPoints()
    self.texture.highlight:SetBlendMode("ADD")
    self.texture.highlight:SetDrawLayer("HIGHLIGHT")
    
    self.frame:SetSize(32, 32)
    self.frame:RegisterForClicks("LeftButtonUp", "RightButtonUp")
    self.frame:SetScript("OnEnter", self.ShowTooltip)
    self.frame:SetScript("OnLeave", self.HideTooltip)
    self.frame:SetScript("OnClick", self.frameclick)
end

function private.objects.ui.textureButton:SetTexture(texture)
    self.texture.normal:SetTexture(texture)
    self.texture.disabled:SetTexture(texture)
end

function private.objects.ui.textureButton:SetDesaturated(des)
    self.texture.disabled:SetDesaturated(des)
end

function private.objects.ui.textureButton:ShowTooltip()
    GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
    GameTooltip:ClearLines()
    GameTooltip:AddLine(self.__parent.name)
    if self.__parent.tooltip then GameTooltip:AddLine(self.__parent.tooltip, 1, 1, 1) end
    GameTooltip:Show()
end

function private.objects.ui.textureButton:HideTooltip()
    GameTooltip:Hide()
end

function private.objects.ui.textureButton:frameclick(...)
    -- DO NOT OVERRIDE (unless you know what you're doing and need to for whatever reason)
    -- This simply pushes the click function from the button to the handler object
    self.__parent:click(...)
end

function private.objects.ui.textureButton:click(button)
    -- Reacts to any click at all
    -- Override this function if you only want one click function for both buttons
    -- Otherwise, override the individual left and right click functions
    if button == "LeftButton" then self:leftclick() end
    if button == "RightButton" then self:rightclick() end
end

function private.objects.ui.textureButton:leftclick()
    -- Reacts to a left click
    -- Override this to handle left clicks manually
    -- If textureButton:click() is overridden without calling this function,
    -- this will not be called on clicks.
    
    -- This is a placeholder function; override it for functionality
end

function private.objects.ui.textureButton:rightclick()
    -- Reacts to a right click
    -- Override this to handle right clicks manually
    -- If textureButton:click() is overridden without calling this function,
    -- this will not be called on clicks.
    
    -- This is a placeholder function; override it for functionality
end

------------------------------------------------------------
----------------------- Spell button -----------------------
------------------------------------------------------------

private.objects.ui.spellButton = private.objects.base()

function private.objects.ui.spellButton:init(parent, spellID, cooldown, reagent)
    self.spellID = spellID
    self.reagent = reagent or 0
    
    self.frame = CreateFrame("Button", nil, parent or UIParent, "SecureActionButtonTemplate")
    self.frame.__parent = self
    self.texture = {}
    
    local spellName = GetSpellInfo(spellID)
    local texture = GetSpellTexture(spellName)
    
    self.texture.normal = self.frame:CreateTexture()
    self.texture.normal:SetTexture(texture)
    self.texture.normal:SetDrawLayer("BACKGROUND")
    self.texture.normal:SetAllPoints()
    
    self.texture.pushed = self.frame:CreateTexture()
    self.texture.pushed:SetTexture("Interface/Buttons/UI-Quickslot-Depress")
    self.texture.pushed:SetAllPoints()
    self.texture.pushed:SetBlendMode("ADD")
    self.texture.pushed:SetDrawLayer("BORDER")
    self.frame:SetPushedTexture(self.texture.pushed)
    
    self.texture.disabled = self.frame:CreateTexture()
    self.texture.disabled:SetTexture(texture)
    self.texture.disabled:SetAllPoints()
    self.texture.disabled:SetDesaturated(true)
    self.frame:SetDisabledTexture(self.texture.disabled)
    
    self.texture.highlight = self.frame:CreateTexture()
    self.texture.highlight:SetTexture("Interface/Buttons/ButtonHilight-Square")
    self.texture.highlight:SetAllPoints()
    self.texture.highlight:SetBlendMode("ADD")
    self.texture.highlight:SetDrawLayer("HIGHLIGHT")
    
    self.label = private.objects.ui.label()
    self.label:init(self.frame, "")
    self.label.frame:SetFontObject(NumberFontNormal)
    self.label.frame:SetPoint("BOTTOMRIGHT")
    if self.reagent == 0 then self.label.frame:Hide() end
    
    self.frame:SetAttribute("type", "spell")
    self.frame:SetAttribute("spell", spellName)
    self.frame:SetAttribute("unit", "player")
    
    self.frame:SetSize(32, 32)
    self.frame:SetScript("OnEnter", self.ShowTooltip)
    self.frame:SetScript("OnLeave", self.HideTooltip)
    
    self.cooldown = CreateFrame("Cooldown", nil, self.frame)
    self.cooldown:SetSwipeTexture('Interface/Cooldown/cooldown2')
    self.cooldown:SetSwipeColor(0, 0, 0, 0.8)
    self.cooldown:SetAllPoints(self.frame)
    self.cooldown:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
    self.cooldown.cooldown = cooldown -- self's cooldown frame has a cooldown attribute, which is set to the cooldown argument.
    self.cooldown.spellID = spellID
    self.cooldown:SetScript("OnEvent", self.spellcastEvent)
end

function private.objects.ui.spellButton:SetSpell(spellID, cooldown, reagent)
    self.spellID = spellID
    self.reagent = reagent or 0
    
    local spellName = GetSpellInfo(spellID)
    local texture = GetSpellTexture(spellName)
    
    self.texture.normal:SetTexture(texture)
    self.texture.disabled:SetTexture(texture)
    
    self.frame:SetAttribute("spell", spellName)
    
    self.cooldown.cooldown = cooldown
    self.cooldown.spellID = spellID
end

function private.objects.ui.spellButton:spellcastEvent(event, unit, spellName, spellRank, _, spellID)
    if unit == "player" and spellID == self.spellID then
        self:SetCooldown(GetTime(), self.cooldown)
    end
end

function private.objects.ui.spellButton:Update()
    -- Re-counts the reagents, and either enables or
    -- disables the button based on the count
    if self.reagent > 0 then
        local count = GetItemCount(self.reagent)
        self.label:SetText(count)
        if count > 0 then
            self.frame:Enable()
        else
            self.frame:Disable()
        end
    end
    
    -- Update the texture
    local spellName = GetSpellInfo(self.spellID)
    local texture = GetSpellTexture(spellName)
    self.texture.normal:SetTexture(texture)
    self.texture.disabled:SetTexture(texture)
    self.frame:SetDisabledTexture(self.texture.disabled)
end

function private.objects.ui.spellButton:CanUse()
    -- Returns whether the player can use this spell or not
    -- (ignoring the cooldown and reagents)
    return IsSpellKnown(self.spellID)
end

function private.objects.ui.spellButton:ShowTooltip()
    GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
    GameTooltip:ClearLines()
    GameTooltip:SetHyperlink(GetSpellLink(self.__parent.spellID))
    GameTooltip:Show()
end

function private.objects.ui.spellButton:HideTooltip()
    GameTooltip:Hide()
end

------------------------------------------------------------
----------------------- Item button ------------------------
------------------------------------------------------------

private.objects.ui.itemButton = private.objects.base()

function private.objects.ui.itemButton:init(parent, itemID, cooldown, spellID, isConsumed)
    self.spellID = spellID
    self.itemID = itemID
    self.isConsumed = isConsumed or false
    
    self.frame = CreateFrame("Button", nil, parent or UIParent, "SecureActionButtonTemplate")
    self.frame.__parent = self
    self.texture = {}
    
    local texture = GetItemIcon(itemID)
    
    self.texture.normal = self.frame:CreateTexture()
    self.texture.normal:SetTexture(texture)
    self.texture.normal:SetDrawLayer("BACKGROUND")
    self.texture.normal:SetAllPoints()
    
    self.texture.pushed = self.frame:CreateTexture()
    self.texture.pushed:SetTexture("Interface/Buttons/UI-Quickslot-Depress")
    self.texture.pushed:SetAllPoints()
    self.texture.pushed:SetBlendMode("ADD")
    self.texture.pushed:SetDrawLayer("BORDER")
    self.frame:SetPushedTexture(self.texture.pushed)
    
    self.texture.disabled = self.frame:CreateTexture()
    self.texture.disabled:SetTexture(texture)
    self.texture.disabled:SetAllPoints()
    self.texture.disabled:SetDesaturated(true)
    self.frame:SetDisabledTexture(self.texture.disabled)
    
    self.texture.highlight = self.frame:CreateTexture()
    self.texture.highlight:SetTexture("Interface/Buttons/ButtonHilight-Square")
    self.texture.highlight:SetAllPoints()
    self.texture.highlight:SetBlendMode("ADD")
    self.texture.highlight:SetDrawLayer("HIGHLIGHT")
    
    self.label = private.objects.ui.label()
    self.label:init(self.frame, "")
    self.label.frame:SetFontObject(NumberFontNormal)
    self.label.frame:SetPoint("BOTTOMRIGHT")
    if not self.isConsumed then self.label.frame:Hide() end
    
    self.frame:SetAttribute("type", "item")
    self.frame:SetAttribute("item", GetItemInfo(itemID))
    self.frame:SetAttribute("unit", "player")
    
    self.frame:SetSize(32, 32)
    self.frame:SetScript("OnEnter", self.ShowTooltip)
    self.frame:SetScript("OnLeave", self.HideTooltip)
    
    self.cooldown = CreateFrame("Cooldown", nil, self.frame)
    self.cooldown:SetSwipeTexture('Interface/Cooldown/cooldown2')
    self.cooldown:SetSwipeColor(0, 0, 0, 0.8)
    self.cooldown:SetAllPoints(self.frame)
    self.cooldown:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
    self.cooldown.cooldown = cooldown -- self's cooldown frame has a cooldown attribute, which is set to the cooldown argument.
    if spellID then
        self.cooldown.spellID = spellID
        self.cooldown:SetScript("OnEvent", self.spellcastEvent)
    end
end

function private.objects.ui.itemButton:SetItem(itemID, cooldown, spellID, isConsumed)
    self.spellID = spellID
    self.itemID = itemID
    self.isConsumed = isConsumed or false
    
    local texture = GetItemIcon(itemID)
    
    self.texture.normal:SetTexture(texture)
    self.texture.disabled:SetTexture(texture)
    
    self.frame:SetAttribute("item", GetItemInfo(itemID))
    
    self.cooldown.cooldown = cooldown -- self's cooldown frame has a cooldown attribute, which is set to the cooldown argument.
    if spellID then
        self.cooldown.spellID = spellID
    end
end

function private.objects.ui.itemButton:spellcastEvent(event, unit, spellName, spellRank, _, spellID)
    if unit == "player" and spellID == self.spellID then
        self:SetCooldown(GetTime(), self.cooldown)
    end
end

function private.objects.ui.itemButton:Update()
    -- Re-counts the reagents, and either enables or
    -- disables the button based on the count
    if self.isConsumed then
        local count = GetItemCount(self.itemID)
        self.label:SetText(count)
        if count > 0 then
            self.frame:Enable()
        else
            self.frame:Disable()
        end
    end
end

function private.objects.ui.itemButton:CanUse()
    -- Returns whether the player can use this item or not
    -- (ignoring the cooldown)
    local count = GetItemCount(self.itemID)
    return count > 0
end

function private.objects.ui.itemButton:ShowTooltip()
    GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
    GameTooltip:ClearLines()
    local _, link = GetItemInfo(self.__parent.itemID)
    GameTooltip:SetHyperlink(link)
    GameTooltip:Show()
end

function private.objects.ui.itemButton:HideTooltip()
    GameTooltip:Hide()
end

------------------------------------------------------------
--------------------------- Bar ----------------------------
------------------------------------------------------------

private.objects.ui.bar = private.objects.base()

local red = { r = 1, g = 0, b = 0 }
local green = { r = 0, g = 1, b = 0 }
local blue = { r = 0, g = 0, b = 1 }
local yellow = { r = 1, g = 1, b = 0 }

function private.objects.ui.bar:init(parent, colour, overfillcolour)
    self.bgframe = CreateFrame("Frame", nil, parent or UIParent)
    self.frame = self.bgframe -- The standard convention is to be able to refer to object.frame
    self.fillframe = CreateFrame("Frame", nil, self.bgframe)
    self.overfillframe = CreateFrame("Frame", nil, self.bgframe)
    
    -- Base measures
    self.bgframe:SetSize(80, 10)
    self.fillframe:SetPoint("LEFT")
    self.fillframe:SetSize(40, 10)
    self.fillframe:Hide()
    self.overfillframe:SetPoint("LEFT", self.fillframe, "RIGHT")
    self.overfillframe:SetSize(40, 10)
    self.overfillframe:Hide()
    self.length = 0
    self.overlength = 0
    
    self:SetColour(colour, overfillcolour)
end

function private.objects.ui.bar:SetColour(colour, overfillcolour)
    -- Takes a table holding keys r, g, b and a.
    
    -- Use green for fill if nothing else was specified
    if not colour then colour = green end
    
    -- Use yellow for overfill if nothing else was specified
    if not overfillcolour then overfillcolour = yellow end
    
    -- Set the background's transparency first
    local t = self.bgframe.texture or self.bgframe:CreateTexture()
    t:SetTexture(0.3, 0.3, 0.3, colour.a or 1)
    t:SetAllPoints(self.bgframe)
    self.bgframe.texture = t
    
    -- Then set the fill bar's colour and transparency
    t = self.fillframe.texture or self.fillframe:CreateTexture()
    t:SetTexture(colour.r, colour.g, colour.b, colour.a or 1)
    t:SetAllPoints(self.fillframe)
    self.fillframe.texture = t
    
    -- Finally, the overfill bar
    t = self.overfillframe.texture or self.overfillframe:CreateTexture()
    t:SetTexture(overfillcolour.r, overfillcolour.g, overfillcolour.b, overfillcolour.a or 1)
    t:SetAllPoints(self.overfillframe)
    self.overfillframe.texture = t
end

function private.objects.ui.bar:SetSize(width, height)
    width = width or self.bgframe:GetWidth()
    height = height or self.bgframe:GetHeight()
    -- Sets the width and height of all three frames
    self.bgframe:SetSize(width, height)
    self.fillframe:SetSize(width, height)
    self.overfillframe:SetSize(width, height)
    self:SetFill(self.length)
    self:SetOverfill(self.overlength)
end

function private.objects.ui.bar:SetFill(length)
    -- Takes a number between 0 and 1
    -- Fills the bar accordingly
    
    -- Numbers outside of that range will be read as either 0 or 1
    length = max(length, 0)
    length = min(length, 1)
    
    -- Calculate the shown overfill
    local overfill = min(self.overlength, 1 - length)
    
    if length > 0 then
        self.fillframe:Show()
        self.fillframe:SetWidth(length * self.bgframe:GetWidth())
        self.overfillframe:ClearAllPoints()
        self.overfillframe:SetPoint("LEFT", self.fillframe, "RIGHT")
    else
        self.fillframe:Hide()
        self.overfillframe:ClearAllPoints()
        self.overfillframe:SetPoint("TOPLEFT", self.frame, "TOPLEFT")
    end
    
    if overfill > 0 then
        self.overfillframe:Show()
        self.overfillframe:SetWidth(overfill * self.bgframe:GetWidth())
    else
        self.overfillframe:Hide()
    end
    
    -- Record the length in case the bar's size changes
    self.length = length
end

function private.objects.ui.bar:SetOverfill(overlength)
    -- Takes a number between 0 and 1
    -- Will show the overfill bar based on that number and the length
    -- of the regular fill
    
    -- Numbers outside of that range will be read as either 0 or 1
    overlength = max(overlength, 0)
    overlength = min(overlength, 1)
    
    -- Calculate the shown overfill
    local overfill = min(overlength, 1 - self.length)
    
    if overfill > 0 then
        self.overfillframe:Show()
        self.overfillframe:SetWidth(overfill * self.bgframe:GetWidth())
    else
        self.overfillframe:Hide()
    end
    
    self.overlength = overlength
end

function private.objects.ui.bar:SetPoint(...)
    self.bgframe:SetPoint(...)
end

------------------------------------------------------------
------------------------- Race bar -------------------------
------------------------------------------------------------

private.objects.ui.raceBar = private.objects.base()

function private.objects.ui.raceBar:init(raceID, parent)
    self.frame = CreateFrame("Button", nil, parent or UIParent)
    self.frame:SetSize(470, 20)
    self.frame.__parent = self
    self.frame:RegisterForClicks("LeftButtonUp", "RightButtonUp")
    self.frame:SetScript("OnEnter", self.ShowTooltip)
    self.frame:SetScript("OnLeave", self.HideTooltip)
    self.frame:SetScript("OnClick", self.frameclick)
    
    self.raceObject = private.objects.data.race:new()
    self.raceObject:init(raceID)
    
    self.raceName = private.objects.ui.label()
    self.raceName:init(self.frame, self.raceObject.name)
    self.raceName.frame:SetPoint("LEFT", 0, 1)
    
    self.overall = private.objects.ui.bar()
    self.overall:init(self.frame)
    self.overall.bgframe:SetPoint("LEFT", 100, 0)
    
    self.overallCount = private.objects.ui.countlabel()
    self.overallCount:init(self.frame)
    self.overallCount.frame:SetPoint("LEFT", 190, 1)
    
    self.current = private.objects.ui.bar()
    self.current:init(self.frame)
    self.current.bgframe:SetPoint("LEFT", 270, 0)
    
    self.currentCount = private.objects.ui.countlabel()
    self.currentCount:init(self.frame)
    self.currentCount.frame:SetPoint("LEFT", 360, 1)
    
    self.keystones = private.objects.ui.countlabel()
    self.keystones:init(self.frame)
    self.keystones.frame:SetPoint("RIGHT", 0, 1)
end

function private.objects.ui.raceBar:update()
    self.raceObject:update()
    
    if self.raceObject.numArtifacts.total > 0 and (self.raceObject.currentArtifact.fragments > 0) then -- Avoid division by 0
        self.overall:SetFill(self.raceObject.completed.total / self.raceObject.numArtifacts.total)
        self.overallCount:SetBoth(self.raceObject.completed.total, self.raceObject.numArtifacts.total)
        
        self.current:SetFill(self.raceObject.regularFragments / self.raceObject.currentArtifact.fragments)
        self.current:SetOverfill(self.raceObject.keystoneFragments / self.raceObject.currentArtifact.fragments)
        self.currentCount:SetBoth(self.raceObject.totalFragments, self.raceObject.currentArtifact.fragments)
    end
    
    if self.raceObject.keystoneID == 0 then
        self.keystones.frame:Hide()
    else
        self.keystones:SetBoth(self.raceObject.numKeystones, self.raceObject.currentArtifact.sockets)
    end
end

function private.objects.ui.raceBar:ShowTooltip()
    GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
    GameTooltip:ClearLines()
    GameTooltip:AddLine(self.__parent.raceObject.name, 1, 1, 1)
    if self.__parent.raceObject.currentArtifact.fragments == 0 then
        GameTooltip:AddLine(L("You have not discovered this race yet!"), 1, 1, 0)
        GameTooltip:Show()
        return
    end
    GameTooltip:AddLine(format(
        L("%d / %d artifacts completed."),
        self.__parent.raceObject.completed.total,
        self.__parent.raceObject.numArtifacts.total
    ))
    GameTooltip:AddLine(format(
        L("%d solves total."),
        self.__parent.raceObject.totalCompleted
    ))
    GameTooltip:AddLine(" ")
    GameTooltip:AddDoubleLine(L("Project") .. ":", self.__parent.raceObject.currentArtifact.name, 1, 1, 0, 1, 1, 1)
    if not self.__parent.raceObject.currentArtifact.artifact then
        -- There's been some error - the current artifact is not registered.
        GameTooltip:AddLine(L("Error: Unknown artifact."), 1, 1, 1)
        GameTooltip:Show()
        return
    end
    GameTooltip:AddLine(format(
        L("Solved %d times"),
        self.__parent.raceObject.currentArtifact.artifact.completed
    ), 1, 1, 1)
    --test = self.__parent.raceObject
    local ready = (self.__parent.raceObject.totalFragments >= self.__parent.raceObject.currentArtifact.fragments)
    GameTooltip:AddDoubleLine(
        format(
            L("%d / %d fragments"),
            self.__parent.raceObject.totalFragments,
            self.__parent.raceObject.currentArtifact.fragments
        ),
        ready and
            L("Ready to solve!")
        or
            format(
                L("Need %d more."),
                self.__parent.raceObject.currentArtifact.fragments - self.__parent.raceObject.totalFragments
            ),
        1, 1, 1, ready and 0 or 1, ready and 1 or 0, 0 -- Yep, this is ugly.
    )
    
    -- If there is a pristine version - show whether it is completed.
    local itemID = self.__parent.raceObject.currentArtifact.artifact.pristineItemID
    if itemID then
        local questID = self.__parent.raceObject.artifacts.pristine:get(itemID)
        if questID then
            local completed = IsQuestFlaggedCompleted(questID)
            local itemName = GetItemInfo(itemID)
            GameTooltip:AddDoubleLine(
                itemName,
                completed and L("Completed") or L("Incomplete"),
                1, 1, 1, completed and 0 or 1, completed and 1 or 0, 0 -- This is ugly too.
            )
        end
    end

    GameTooltip:AddLine(" ")
    GameTooltip:AddLine(L("Click to open artifact list."), 0, 1, 0)
    if ready then GameTooltip:AddLine(L("Right click to solve."), 0, 1, 0) end
    
    GameTooltip:Show()
end

function private.objects.ui.raceBar:HideTooltip()
    GameTooltip:Hide()
end

function private.objects.ui.raceBar:frameclick(...)
    -- DO NOT OVERRIDE (unless you know what you're doing and need to for whatever reason)
    -- This simply pushes the click function from the button to the handler object
    self.__parent:click(...)
end

function private.objects.ui.raceBar:click(button)
    -- Reacts to any click at all
    -- Override this function if you only want one click function for both buttons
    -- Otherwise, override the individual left and right click functions
    if button == "LeftButton" then self:leftclick() end
    if button == "RightButton" then self:rightclick() end
end

function private.objects.ui.raceBar:leftclick()
    -- Reacts to a left click
    -- Override this to handle left clicks manually
    -- If textureButton:click() is overridden without calling this function,
    -- this will not be called on clicks.
    
    -- This is a placeholder function; override it for functionality
end

function private.objects.ui.raceBar:rightclick()
    -- Reacts to a right click
    -- Override this to handle right clicks manually
    -- If textureButton:click() is overridden without calling this function,
    -- this will not be called on clicks.
    
    -- Solve the artifact
    self.raceObject:Solve()
end

------------------------------------------------------------
--------------------- Achievement bar ----------------------
------------------------------------------------------------

private.objects.ui.achievementBar = private.objects.base()

function private.objects.ui.achievementBar:init(parent)
    -- Set up the frame
    self.frame = CreateFrame("Button", nil, parent or UIParent)
    self.frame:SetSize(470, 20)
    self.frame.__parent = self
    
    self.frame:SetScript("OnEnter", self.ShowTooltip)
    self.frame:SetScript("OnLeave", self.HideTooltip)
    
    self.name = private.objects.ui.label()
    self.name:init(self.frame, "")
    self.name.frame:SetPoint("LEFT")
    
    self.bar = private.objects.ui.bar()
    self.bar:init(self.frame)
    self.bar.frame:SetPoint("LEFT", 300, 0)
    self.bar:SetFill(0.5)
    
    self.count = private.objects.ui.countlabel()
    self.count:init(self.frame)
    self.count.frame:SetPoint("LEFT", 400, 0)
    
    self.achievementObjects = list()
end

function private.objects.ui.achievementBar:AddAchievement(achievementObject)
    self.achievementObjects:add(achievementObject)
    achievementObject:SetFrame(self)
    self:update()
end

function private.objects.ui.achievementBar:update()
    -- Toss out all completed achievements (except the last one)
    while (self.achievementObjects[1]:getProgress() == 1) and (#self.achievementObjects > 1) do
        local object = self.achievementObjects[1]
        self.achievementObjects:remove(object)
    end
    
    -- Set the achievement name
    self.name:SetText(self.achievementObjects[1].name)
    
    -- Update the bar
    self.bar:SetFill(self.achievementObjects[1]:getProgress())
    self.count:SetBoth(self.achievementObjects[1]:getNumericProgress())
end

function private.objects.ui.achievementBar:ShowTooltip()
    GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
    GameTooltip:ClearLines()
    local lines = self.__parent.achievementObjects[1]:getLines()
    GameTooltip:AddLine(self.__parent.achievementObjects[1].name)
    for i = 1, #lines do
        GameTooltip:AddLine(lines[i], 1, 1, 1)
    end
    GameTooltip:Show()
end

function private.objects.ui.achievementBar:HideTooltip()
    GameTooltip:Hide()
end

------------------------------------------------------------
---------------------- Frame backdrop ----------------------
------------------------------------------------------------

local frameBackdrop = {
    bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
    edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
    tile = true, tileSize = 16, edgeSize = 16,
    insets = { left = 5, right = 5, top = 5, bottom = 5 } }

------------------------------------------------------------
---------------------- Artifact list -----------------------
------------------------------------------------------------

local function commonSort(a, b)
    return a[2] > b[2]
end

private.objects.ui.artifactList = private.objects.base()

function private.objects.ui.artifactList:init(parent)
    self.defaults = {}
    self.defaults.numLines = 30
    
    self.allowScroll = false
    self.offset = 0
    
    -- Set up the frame
    self.frame = CreateFrame("Frame", nil, parent)
    self.frame.__parent = self
    self.frame:SetBackdrop(frameBackdrop)
    self.frame:SetSize(400, 490)
    self.frame:SetPoint("TOPLEFT", parent, "TOPRIGHT")
    self.frame:SetScript("OnMouseWheel", self.ScrollHandler)
    
    self.slider = CreateFrame("Slider", "artifactframe" .. math.random(), self.frame, "OptionsSliderTemplate")
    self.slider.__parent = self
    self.slider:SetWidth(10)
    self.slider:SetHeight(500)
    self.slider:SetOrientation("VERTICAL")
    self.slider:SetPoint("RIGHT")
    self.slider:Enable()
    self.slider:SetMinMaxValues(0, 0)
    self.slider:SetValueStep(1)
    self.slider:SetValue(0)
    self.slider:SetScript("OnValueChanged", self.SlideHandler)
    
    -- Remove the darn text
    _G[self.slider:GetName() .. "Low"]:SetText("")
    _G[self.slider:GetName() .. "High"]:SetText("")
    
    self.lines = list()
    self.links = {}
    self.labels = {}
    
    for i=1, 30 do
        self.links[i] = private.objects.ui.itemLinkLabel()
        self.links[i]:init(self.frame, 0)
        self.links[i].frame:SetPoint("TOPLEFT", 10, -15 - 15 * i)
        
        self.labels[i] = private.objects.ui.label()
        self.labels[i]:init(self.frame, " ")
        self.labels[i].frame:SetPoint("TOPRIGHT", -15, -15 - 15 * i)
    end
    
    -- Close button
    self.closeButton = CreateFrame("Button", nil, self.frame, "UIPanelCloseButton")
    self.closeButton:SetPoint("TOPRIGHT", -10, -10)
    
    self.frame:Hide()
end

function private.objects.ui.artifactList:PopulateRace(raceBar)
    -- Populate the artifact list with the artifacts for a given race
    
    -- Clear out the line list
    self.lines:clear()
    
    -- Add the rare header
    self.lines:add({ "|c00FFFF00" .. L("RARE ARTIFACTS") .. "|r", " " })
    
    -- Read the rare artifacts
    local artifacts = raceBar.raceObject.artifacts
    for k, artifact in pairs(artifacts.rare) do
        self.lines:add({ artifact.itemID, artifact.completed })
    end
    
    -- If the race has pristine artifacts
    if raceBar.raceObject.numArtifacts.pristine > 0 then
        -- Add the pristine header
        self.lines:add({ " ", " " })
        self.lines:add({ "|c00FFFF00" .. L("PRISTINE ARTIFACTS") .. "|r", " " })
        
        -- Add the pristine artifacts
        for itemID, questID in pairs(artifacts.pristine) do
            local completed = IsQuestFlaggedCompleted(questID)
            self.lines:add({ itemID, completed and 1 or 0 })
        end
    end
    
    -- Add the common header
    self.lines:add({ " ", " " })
    self.lines:add({ "|c00FFFF00" .. L("COMMON ARTIFACTS") .. "|r", " " })
    
    -- Read the common artifacts
    local commons = list()
    for k, artifact in pairs(artifacts.common) do
        commons:add({ artifact.itemID, artifact.completed })
    end
    
    -- Sort the common artifacts
    table.sort(commons, commonSort)
    
    -- Merge the common artifacts back into the list
    self.lines:merge(commons)
    
    -- Prepare the slider
    self.offset = 0
    --if #self.leftlines > self.defaults.numLines then
    if #self.lines > self.defaults.numLines then
        local overflow = #self.lines - self.defaults.numLines
        self.slider:SetMinMaxValues(0, overflow)
        self.slider:SetValue(0)
        self.slider:Show()
        self.allowScroll = true
    else
        self.slider:Hide()
        self.allowScroll = false
    end
    
    -- Update the list
    self:UpdateList(0)
end

function private.objects.ui.artifactList:PopulateHave(raceBarList, includeRare, includeCommon)
    -- Populate the artifact list with all artifacts the player has collected
    
    -- Clear out the line list
    self.lines:clear()
    
    if includeRare then
        -- Header
        self.lines:add({ "|c00FFFF00" .. L("RARE ARTIFACTS") .. "|r", " " })
        
        for i = 1, #raceBarList do
            local artifacts = raceBarList[i].raceObject.artifacts
            for k, artifact in pairs(artifacts.rare) do
                if artifact.completed > 0 then
                    self.lines:add({ artifact.itemID, artifact.completed })
                end
            end
        end
    end
    
    if includeCommon then
        -- Header
        self.lines:add({ " ", " " })
        self.lines:add({ "|c00FFFF00" .. L("COMMON ARTIFACTS") .. "|r", " " })
        
        local commons = list()
        
        for i = 1, #raceBarList do
            local artifacts = raceBarList[i].raceObject.artifacts
            for k, artifact in pairs(artifacts.common) do
                if artifact.completed > 0 then
                    commons:add({ artifact.itemID, artifact.completed })
                end
            end
        end
        
        -- Sort the common artifacts
        table.sort(commons, commonSort)
        
        -- Merge the common artifacts into the list
        self.lines:merge(commons)
    end
    
    -- Prepare the slider
    self.offset = 0
    --if #self.leftlines > self.defaults.numLines then
    if #self.lines > self.defaults.numLines then
        local overflow = #self.lines - self.defaults.numLines
        self.slider:SetMinMaxValues(0, overflow)
        self.slider:SetValue(0)
        self.slider:Show()
        self.allowScroll = true
    else
        self.slider:Hide()
        self.allowScroll = false
    end
    
    -- Update the list
    self:UpdateList(0)
end

function private.objects.ui.artifactList:PopulateMissing(raceBarList, includeRare, includeCommon)
    -- Populate the artifact list with all artifacts the player has collected
    
    -- Clear out the line list
    self.lines:clear()
    
    if includeRare then
        -- Header
        self.lines:add({ "|c00FFFF00" .. L("RARE ARTIFACTS") .. "|r", " " })
        
        for i = 1, #raceBarList do
            local artifacts = raceBarList[i].raceObject.artifacts
            for k, artifact in pairs(artifacts.rare) do
                if artifact.completed == 0 then
                    self.lines:add({ artifact.itemID, artifact.completed })
                end
            end
        end
    end
    
    if includeCommon then
        -- Header
        self.lines:add({ " ", " " })
        self.lines:add({ "|c00FFFF00" .. L("COMMON ARTIFACTS") .. "|r", " " })
        
        local commons = list()
        
        for i = 1, #raceBarList do
            local artifacts = raceBarList[i].raceObject.artifacts
            for k, artifact in pairs(artifacts.common) do
                if artifact.completed == 0 then
                    commons:add({ artifact.itemID, artifact.completed })
                end
            end
        end
        
        -- Sort the common artifacts
        table.sort(commons, commonSort)
        
        -- Merge the common artifacts into the list
        self.lines:merge(commons)
    end
    
    -- Prepare the slider
    self.offset = 0
    --if #self.leftlines > self.defaults.numLines then
    if #self.lines > self.defaults.numLines then
        local overflow = #self.lines - self.defaults.numLines
        self.slider:SetMinMaxValues(0, overflow)
        self.slider:SetValue(0)
        self.slider:Show()
        self.allowScroll = true
    else
        self.slider:Hide()
        self.allowScroll = false
    end
    
    -- Update the list
    self:UpdateList(0)
end

function private.objects.ui.artifactList:PopulateAll(raceBarList, includeRare, includeCommon)
    -- Populate the artifact list with all artifacts
    
    -- Clear out the line list
    self.lines:clear()
    
    if includeRare then
        -- Header
        self.lines:add({ "|c00FFFF00" .. L("RARE ARTIFACTS") .. "|r", " " })
        
        for i = 1, #raceBarList do
            local artifacts = raceBarList[i].raceObject.artifacts
            for k, artifact in pairs(artifacts.rare) do
                self.lines:add({ artifact.itemID, artifact.completed })
            end
        end
    end
    
    if includeCommon then
        -- Header
        self.lines:add({ " ", " " })
        self.lines:add({ "|c00FFFF00" .. L("COMMON ARTIFACTS") .. "|r", " " })
        
        local commons = list()
        
        for i = 1, #raceBarList do
            local artifacts = raceBarList[i].raceObject.artifacts
            for k, artifact in pairs(artifacts.common) do
                commons:add({ artifact.itemID, artifact.completed })
            end
        end
        
        -- Sort the common artifacts
        table.sort(commons, commonSort)
        
        -- Merge the common artifacts into the list
        self.lines:merge(commons)
    end
    
    -- Prepare the slider
    self.offset = 0
    --if #self.leftlines > self.defaults.numLines then
    if #self.lines > self.defaults.numLines then
        local overflow = #self.lines - self.defaults.numLines
        self.slider:SetMinMaxValues(0, overflow)
        self.slider:SetValue(0)
        self.slider:Show()
        self.allowScroll = true
    else
        self.slider:Hide()
        self.allowScroll = false
    end
    
    -- Update the list
    self:UpdateList(0)
end

function private.objects.ui.artifactList:UpdateList(offset)
    -- This is called to update where the lines show in the frame.
    -- Called at the end of every populate function, and also
    -- whenever the user scrolls.
    
    self.frame:Show()
    self.offset = offset
    
    for i = 1, self.defaults.numLines do
        if self.lines[i + offset] then
            self.links[i]:SetItem(self.lines[i + offset][1]) -- No need for a check, the link object handles that
            local right = tonumber(self.lines[i + offset][2])
            if right then
                if right > 1 then
                    self.labels[i]:SetText("|c0000FF00[" .. right .. " " .. L("completed") .. "]|r")
                elseif right == 1 then
                    self.labels[i]:SetText("|c0000FF00[" .. L("Completed") .. "]|r")
                else
                    self.labels[i]:SetText("|c00FF0000[" .. L("Incomplete") .. "]|r")
                end
            else
                self.labels[i]:SetText(" ")
            end
        else
            self.links[i]:SetText(" ")
            self.labels[i]:SetText(" ")
        end
    end
end

function private.objects.ui.artifactList:SlideHandler(value)
    if self.__parent.allowScroll then
        self.__parent:UpdateList(value)
    end
end

function private.objects.ui.artifactList:ScrollHandler(delta)
    if self.__parent.allowScroll then
        -- Simply setting the slider's value allows the slider to handle
        -- min/max values, instead of me doing it a second time.
        self.__parent.slider:SetValue(self.__parent.offset - delta)
    end
end

------------------------------------------------------------
------------------------ Side frame ------------------------
------------------------------------------------------------

local sideFrameList = list()

private.objects.ui.sideFrame = private.objects.base()

function private.objects.ui.sideFrame:init(parent)
    -- Set up the frame
    self.frame = CreateFrame("Frame", nil, parent or UIParent)
    self.frame:SetBackdrop(frameBackdrop)
    self.frame:Hide()
    
    self.buttons = dict()
    
    self.height = 8 -- The maximum height for this frame (measured in buttons)
    
    -- Add self to the side frame list
    sideFrameList:add(self)
end

function private.objects.ui.sideFrame:addSpell(spellID, cooldown, reagent)
    -- Creates a spell button and adds it to this frame's list of buttons
    if self.buttons:get(spellID) then return end -- Don't add a spell more than once.
    local spellButton = private.objects.ui.spellButton()
    spellButton:init(self.frame, spellID, cooldown, reagent)
    self.buttons:set(spellID, spellButton)
end

function private.objects.ui.sideFrame:addItem(itemID, spellID, cooldown, consumed)
    -- Creates an item button and adds it to this frame's list of buttons
    if self.buttons:get(itemID) then return end -- Don't add an item more than once.
    local itemButton = private.objects.ui.itemButton()
    itemButton:init(self.frame, itemID, cooldown, spellID, consumed)
    self.buttons:set(itemID, itemButton)
end

function private.objects.ui.sideFrame:Show()
    -- Hides all other side frames, and shows this one
    for k, v in pairs(sideFrameList) do
        v:Hide()
    end
    
    -- Decide which buttons to display
    local i = 0
    for k, v in pairs(self.buttons) do
        v:Update()
        if v:CanUse() then
            v.frame:SetPoint("TOPRIGHT", -8 - 36 * math.floor(i / self.height), -8 - 36 * (i % self.height))
            v.frame:Show()
            i = i + 1
        else
            v.frame:Hide()
        end
    end
    
    if i > 0 then -- Don't show the frame if there are no buttons!
        self.frame:SetSize(12 + 36 * math.ceil(i / self.height), 12 + 36 * min(i, self.height))
        self.frame:Show()
    end
end

function private.objects.ui.sideFrame:Hide()
    -- Hides this side frame
    self.frame:Hide()
end

function private.objects.ui.sideFrame:Toggle()
    -- Shows this frame if it is not shown, hides it if it is
    if self.frame:IsShown() then self:Hide() else self:Show() end
end

------------------------------------------------------------
-------------------- Achievement frame ---------------------
------------------------------------------------------------

private.objects.ui.achievements = private.objects.base()

function private.objects.ui.achievements:init(parent)
    self.page = 1
    
    -- Set up the frame
    self.frame = CreateFrame("Frame", nil, parent or UIParent)
    self.frame:SetBackdrop(frameBackdrop)
    
    self.frame:SetSize(500, 275)
    self.frame:SetPoint("TOPLEFT", parent, "BOTTOMLEFT")
    
    -- Set up the achievement bars
    self.bars = list()
    for i = 1, #private.achievementOrder do
        local bar = private.objects.ui.achievementBar()
        bar:init(self.frame)
        bar.frame:SetPoint("TOPLEFT", 15, -15 - 20 * (((i - 1) % 10) + 1))
        self.bars:add(bar)
        
        local orderData = private.achievementOrder[i]
        if type(orderData) == "table" then
            for k, v in pairs(orderData) do
                bar:AddAchievement(private.achievements[v])
            end
        else
            bar:AddAchievement(private.achievements[orderData])
        end
        
        if i > 10 then bar.frame:Hide() end
    end
    
    -- Create the next/previous page buttons
    self.prev = CreateFrame("Button", nil, self.frame)
    self.prev:SetNormalTexture("Interface/Buttons/UI-SpellbookIcon-PrevPage-Up")
    self.prev:SetPushedTexture("Interface/Buttons/UI-SpellbookIcon-PrevPage-Down")
    self.prev:SetDisabledTexture("Interface/Buttons/UI-SpellbookIcon-PrevPage-Disabled")
    self.prev:SetHighlightTexture("Interface/Buttons/ButtonHilight-Square")
    self.prev:SetSize(32, 32)
    self.prev:SetPoint("BOTTOMLEFT", 10, 10)
    self.prev:Disable()
    self.prev.__parent = self
    self.prev:RegisterForClicks("LeftButtonUp")
    self.prev:SetScript("OnClick", function(self) self.__parent:PrevPage() end)
    
    self.next = CreateFrame("Button", nil, self.frame)
    self.next:SetNormalTexture("Interface/Buttons/UI-SpellbookIcon-NextPage-Up")
    self.next:SetPushedTexture("Interface/Buttons/UI-SpellbookIcon-NextPage-Down")
    self.next:SetDisabledTexture("Interface/Buttons/UI-SpellbookIcon-NextPage-Disabled")
    self.next:SetHighlightTexture("Interface/Buttons/ButtonHilight-Square")
    self.next:SetSize(32, 32)
    self.next:SetPoint("BOTTOMRIGHT", -10, 10)
    self.next.__parent = self
    self.next:RegisterForClicks("LeftButtonUp")
    self.next:SetScript("OnClick", function(self) self.__parent:NextPage() end)
    
    -- Create the page number label
    self.pageLabel = private.objects.ui.label()
    self.pageLabel:init(self.frame, format(L("Page %d / %d"), 1, math.ceil(#self.bars / 10)), 12)
    self.pageLabel.frame:SetPoint("BOTTOM", 0, 15)
    
    -- Create the close button
    self.close = CreateFrame("Button", nil, self.frame, "UIPanelCloseButton")
    self.close:SetPoint("TOPRIGHT", -5, -5)
    
    self.frame:Hide()
end

function private.objects.ui.achievements:Show()
    -- Updates all the achievement bars
    for i=1, #self.bars do
        self.bars[i]:update()
    end
    self.frame:Show()
end

function private.objects.ui.achievements:Hide()
    self.frame:Hide()
end

function private.objects.ui.achievements:Toggle()
    if self.frame:IsShown() then self:Hide() else self:Show() end
end

function private.objects.ui.achievements:ShowPage(pageNum)
    -- Show the bars corresponding to the requested page
    if pageNum > math.ceil(#self.bars / 10) then return end
    if pageNum < 1 then return end
    for i=1, #self.bars do
        self.bars[i].frame:Hide()
    end
    for i=1, 10 do
        if #self.bars >= i + ((pageNum - 1) * 10) then
            self.bars[i + ((pageNum - 1) * 10)].frame:Show()
        end
    end
    
    -- Update the page number label
    self.pageLabel:SetText(format(L("Page %d / %d"), pageNum, math.ceil(#self.bars / 10)))
    
    self.page = pageNum
end

function private.objects.ui.achievements:PrevPage()
    -- Turn to the previous page
    self:ShowPage(self.page - 1)
    
    -- If we're on (or before) the first page, disable the prev button
    if self.page <= 1 then
        self.prev:Disable()
    else
        self.prev:Enable()
    end
    
    -- If we're on (or beyond) the last page, disable the next button
    if self.page * 10 >= #self.bars then
        self.next:Disable()
    else
        self.next:Enable()
    end
end

function private.objects.ui.achievements:NextPage()
    -- Turn to the next page
    self:ShowPage(self.page + 1)
    
    -- If we're on (or before) the first page, disable the prev button
    if self.page <= 1 then
        self.prev:Disable()
    else
        self.prev:Enable()
    end
    
    -- If we're on (or beyond) the last page, disable the next button
    if self.page * 10 >= #self.bars then
        self.next:Disable()
    else
        self.next:Enable()
    end
end

------------------------------------------------------------
----------------------- Links frame ------------------------
------------------------------------------------------------

private.objects.ui.linksFrame = private.objects.base()

function private.objects.ui.linksFrame:init(parent)
    -- Set up the frame
    self.frame = CreateFrame("Frame", nil, parent or UIParent)
    self.frame:SetBackdrop(frameBackdrop)
    self.frame:SetSize(500, 265)
    self.frame:SetPoint("TOPLEFT", parent, "BOTTOMLEFT")
    
    self.projectList = list()
    self.itemList = list()
    self.linkList = list()
    
    self.selectedProject = ""
    self.selectedItem = 0
    self.selectedItemName = ""
    self.selectedRace = 0
    
    self.projectFrame = CreateFrame("Frame", nil, self.frame)
    self.projectFrame:SetSize(500, 265)
    self.projectFrame:SetPoint("TOPLEFT")
    self.projectFrame.buttons = {}
    self.projectFrame.header = private.objects.ui.label()
    self.projectFrame.header:init(self.projectFrame, "|c00FFFF00" .. L("Select a project") .. "|r", 16)
    self.projectFrame.header.frame:SetPoint("TOP", 0, -10)
    
    self.itemFrame = CreateFrame("Frame", nil, self.frame)
    self.itemFrame:SetSize(500, 265)
    self.itemFrame:SetPoint("TOPLEFT")
    self.itemFrame.buttons = {}
    self.itemFrame.header = private.objects.ui.label()
    self.itemFrame.header:init(self.itemFrame, "|c00FFFF00" .. L("Select an item") .. "|r", 16)
    self.itemFrame.header.frame:SetPoint("TOP", 0, -10)
    self.itemFrame:Hide()
    
    -- Link frame
    self.linkFrame = CreateFrame("Frame", nil, self.frame)
    self.linkFrame:SetSize(500, 265)
    self.linkFrame:SetPoint("TOPLEFT")
    self.linkFrame.buttons = {}
    self.linkFrame.header = private.objects.ui.label()
    self.linkFrame.header:init(self.linkFrame, "|c00FFFF00" .. L("Link?") .. "|r", 16)
    self.linkFrame.header.frame:SetPoint("TOP", 0, -10)
    
    -- Link frame subframes
    self.linkFrame.project = CreateFrame("Frame", nil, self.linkFrame)
    self.linkFrame.project:SetSize(470, 70)
    self.linkFrame.project:SetPoint("TOP", 0, -40)
    self.linkFrame.project.header = private.objects.ui.label()
    self.linkFrame.project.header:init(self.linkFrame.project, L("Project"), 14)
    self.linkFrame.project.header.frame:SetPoint("TOPLEFT", 10, -10)
    self.linkFrame.project.button = private.objects.ui.textureButton()
    self.linkFrame.project.button:init(self.linkFrame.project, "INTERFACE/ICONS/inv_misc_shovel_01", "", "")
    self.linkFrame.project.button:SetDesaturated(false)
    self.linkFrame.project.button.frame:Disable()
    self.linkFrame.project.button.frame:SetPoint("TOPLEFT", 10, -30)
    self.linkFrame.project.name = private.objects.ui.label()
    self.linkFrame.project.name:init(self.linkFrame.project, "")
    self.linkFrame.project.name.frame:SetPoint("TOPLEFT", 46, -40)
    
    self.linkFrame.item = CreateFrame("Frame", nil, self.linkFrame)
    self.linkFrame.item:SetSize(470, 70)
    self.linkFrame.item:SetPoint("TOP", 0, -120)
    self.linkFrame.item.header = private.objects.ui.label()
    self.linkFrame.item.header:init(self.linkFrame.item, L("Item"), 14)
    self.linkFrame.item.header.frame:SetPoint("TOPLEFT", 10, -10)
    self.linkFrame.item.button = private.objects.ui.textureButton()
    self.linkFrame.item.button:init(self.linkFrame.item, "INTERFACE/ICONS/inv_misc_shovel_01", "", "")
    self.linkFrame.item.button:SetDesaturated(false)
    self.linkFrame.item.button.frame:Disable()
    self.linkFrame.item.button.frame:SetPoint("TOPLEFT", 10, -30)
    self.linkFrame.item.name = private.objects.ui.label()
    self.linkFrame.item.name:init(self.linkFrame.item, "")
    self.linkFrame.item.name.frame:SetPoint("TOPLEFT", 46, -40)
    
    self.linkFrame.linkButton = CreateFrame("Button", nil, self.linkFrame, "OptionsButtonTemplate")
    self.linkFrame.linkButton.__parent = self
    self.linkFrame.linkButton:SetText(L("Link!"))
    self.linkFrame.linkButton:SetPoint("TOP", 0, -210)
    self.linkFrame.linkButton:SetScript("OnClick", function(self)
        -- Add the link
        private.links:addLink(self.__parent.selectedItemName, self.__parent.selectedProject)
        
        -- Update the race
        self.__parent.selectedRace:update()
        
        -- Hide and show the frame
        self.__parent:Hide()
        self.__parent:Show()
    end)
    
    self.linkFrame:Hide()
    
    -- Unlink frame button
    self.unlinkButton = private.objects.ui.textureButton()
    self.unlinkButton.__parent = self
    self.unlinkButton:init(self.frame, "Interface/ICONS/Ability_ThunderClap", "Remove link", L("Click to remove an existing link"))
    self.unlinkButton.frame:SetPoint("BOTTOMRIGHT", -8, 8)
    function self.unlinkButton:leftclick(...)
        -- Rebuild the list of links (by project)
        self.__parent.linkList:clear()
        for k, v in pairs(private.links.byProject) do
            self.__parent.linkList:add(k)
        end
        
        self.__parent:hideSubFrames()
        self.__parent.linkListFrame:Show()
        
        self.__parent:scrollLinkList(0)
        
        self.__parent.relinkButton.frame:Show()
        self.frame:Hide()
    end
    
    -- Return to link frame button
    self.relinkButton = private.objects.ui.textureButton()
    self.relinkButton.__parent = self
    self.relinkButton:init(self.frame, "Interface/Icons/INV_Belt_18", "Back to linking", L("Click to go back to the project list"))
    self.relinkButton.frame:SetPoint("BOTTOMRIGHT", -8, 8)
    function self.relinkButton:leftclick(...)
        self.__parent:update()
    end
    
    -- Link list frame
    self.linkListFrame = CreateFrame("Frame", nil, self.frame)
    self.linkListFrame:SetSize(500, 265)
    self.linkListFrame:SetPoint("TOPLEFT")
    self.linkListFrame.buttons = {}
    self.linkListFrame.header = private.objects.ui.label()
    self.linkListFrame.header:init(self.linkListFrame, "|c00FFFF00" .. L("Click on a link to remove it") .. "|r", 16)
    self.linkListFrame.header.frame:SetPoint("TOP", 0, -10)
    self.linkListFrame:Hide()
    
    local function projectClick(self)
        self.__parent.selectedProject = self.value
        self.__parent.projectFrame:Hide()
        self.__parent.itemFrame:Show()
        
        -- Populate the item window
        local data = private.links.unknownProjects:get(self.value)
        local race = private.raceList:getRace(data[1])
        self.__parent.selectedRace = race
        local rarity = data[2] == 1 and "rare" or "common"
        local artifacts = race.artifacts[rarity]
        
        self.__parent.itemList:clear()
        for itemName, artifact in pairs(artifacts) do
            if artifact.completed == 0 then
                self.__parent.itemList:add(artifact.itemID)
            end
        end
        
        self.__parent:scrollItemList(0)
    end
    
    local function itemClick(self)
        self.__parent.selectedItem = self.value
        self.__parent.itemFrame:Hide()
        self.__parent.linkFrame:Show()
        
        -- Populate the link window
        local projectData = private.links.unknownProjects:get(self.__parent.selectedProject)
        
        self.__parent.linkFrame.project.name:SetText(self.__parent.selectedProject)
        self.__parent.linkFrame.project.button:SetTexture(projectData[3])
        
        local itemName, _, _, _, _, _, _, _, _, icon = GetItemInfo(self.value)
        self.__parent.selectedItemName = itemName
        self.__parent.linkFrame.item.name:SetText(itemName)
        self.__parent.linkFrame.item.button:SetTexture(icon)
    end
    
    local function linkClick(self)
        private.links:removeLink(self.value)
        
        -- Rebuild the list of links (by project)
        self.__parent.linkList:clear()
        for k, v in pairs(private.links.byProject) do
            self.__parent.linkList:add(k)
        end
        
        -- Update the project list
        self.__parent.projectList:clear()
        for k, v in pairs(private.links.unknownProjects) do
            self.__parent.projectList:add(k)
        end
        
        self.__parent:scrollLinkList(0)
    end
    
    -- Buttons
    for i=1, 15 do
        self.projectFrame.buttons[i] = CreateFrame("Button", nil, self.projectFrame)
        self.projectFrame.buttons[i].__parent = self
        self.projectFrame.buttons[i].value = ""
        self.projectFrame.buttons[i]:SetSize(450, 15)
        self.projectFrame.buttons[i]:SetPoint("TOPLEFT", 10, -15 - 15 * i)
        self.projectFrame.buttons[i]:SetNormalFontObject('GameFontHighlightLeft')
        self.projectFrame.buttons[i]:SetScript("OnClick", projectClick)
        
        self.itemFrame.buttons[i] = CreateFrame("Button", nil, self.itemFrame)
        self.itemFrame.buttons[i].__parent = self
        self.itemFrame.buttons[i].value = 0
        self.itemFrame.buttons[i]:SetSize(450, 15)
        self.itemFrame.buttons[i]:SetPoint("TOPLEFT", 10, -15 - 15 * i)
        self.itemFrame.buttons[i]:SetNormalFontObject('GameFontHighlightLeft')
        self.itemFrame.buttons[i]:SetScript("OnClick", itemClick)
        
        self.linkListFrame.buttons[i] = CreateFrame("Button", nil, self.linkListFrame)
        self.linkListFrame.buttons[i].__parent = self
        self.linkListFrame.buttons[i].value = ""
        self.linkListFrame.buttons[i]:SetSize(450, 15)
        self.linkListFrame.buttons[i]:SetPoint("TOPLEFT", 10, -15 - 15 * i)
        self.linkListFrame.buttons[i]:SetNormalFontObject('GameFontHighlightLeft')
        self.linkListFrame.buttons[i]:SetScript("OnClick", linkClick)
    end
    
    self:update()
    
    self.frame:Hide()
end

function private.objects.ui.linksFrame:update()
    -- Update the project list
    self.projectList:clear()
    for k, v in pairs(private.links.unknownProjects) do
        self.projectList:add(k)
    end
    
    self:hideSubFrames()
    self.projectFrame:Show()
    
    self.unlinkButton.frame:Show()
    self.relinkButton.frame:Hide()
    
    self:scrollProjectList(0)
end

function private.objects.ui.linksFrame:hideSubFrames()
    self.itemFrame:Hide()
    self.projectFrame:Hide()
    self.linkFrame:Hide()
    self.linkListFrame:Hide()
end

--[[
These functions would suggest that the links frame is capable of scrolling,
however it is not. If it at any point becomes necessary to allow this window
to scroll, I will add the necessary code.
]]
function private.objects.ui.linksFrame:scrollProjectList(offset)
    for i=1, 15 do
        if #self.projectList >= i + offset then
            self.projectFrame.buttons[i]:SetText(self.projectList[i + offset])
            self.projectFrame.buttons[i].value = self.projectList[i + offset]
            self.projectFrame.buttons[i]:Enable()
        else
            self.projectFrame.buttons[i]:SetText("")
            self.projectFrame.buttons[i].value = ""
            self.projectFrame.buttons[i]:Disable()
        end
    end
end

function private.objects.ui.linksFrame:scrollItemList(offset)
    for i=1, 15 do
        if #self.itemList >= i + offset then
            local itemName = GetItemInfo(self.itemList[i + offset])
            self.itemFrame.buttons[i]:SetText(itemName)
            self.itemFrame.buttons[i].value = self.itemList[i + offset]
            self.itemFrame.buttons[i]:Enable()
        else
            self.itemFrame.buttons[i]:SetText("")
            self.itemFrame.buttons[i].value = ""
            self.itemFrame.buttons[i]:Disable()
        end
    end
end

function private.objects.ui.linksFrame:scrollLinkList(offset)
    for i=1, 15 do
        if #self.linkList >= i + offset then
            local itemName = private.links:getItem(self.linkList[i])
            self.linkListFrame.buttons[i]:SetText(self.linkList[i] .. " <-> " .. itemName)
            self.linkListFrame.buttons[i].value = self.linkList[i]
            self.linkListFrame.buttons[i]:Enable()
        else
            self.linkListFrame.buttons[i]:SetText("")
            self.linkListFrame.buttons[i].value = ""
            self.linkListFrame.buttons[i]:Disable()
        end
    end
end

function private.objects.ui.linksFrame:Show()
    self:update()
    self.frame:Show()
end

function private.objects.ui.linksFrame:Hide()
    self.frame:Hide()
end

function private.objects.ui.linksFrame:Toggle()
    if self.frame:IsShown() then
        self:Hide()
    else
        self:Show()
    end
end

------------------------------------------------------------
------------------------ Race frame ------------------------
------------------------------------------------------------

private.objects.ui.raceFrame = private.objects.base()

function private.objects.ui.raceFrame:init(parent)
    -- Set up the frame
    self.frame = CreateFrame("Frame", nil, parent or UIParent)
    self.frame:SetBackdrop(frameBackdrop)
    
    self.frame:EnableMouse(true)
    self.frame:SetMovable(true)
    self.frame:SetClampedToScreen(true)
    self.frame:SetScript("OnDragStart", self.frame.StartMoving)
    self.frame:SetScript("OnDragStop", self.frame.StopMovingOrSizing)
    self.frame:RegisterForDrag("LeftButton")
    self.frame:SetWidth(500)
    self.frame:SetPoint("CENTER")
    
    self.skillBar = private.objects.ui.bar()
    self.skillBar:init(self.frame)
    self.skillBar.frame:SetPoint("TOPLEFT", 65, -16)
    self.skillBar:SetSize(300, 10)
    
    self.skillBar.label = private.objects.ui.label()
    self.skillBar.label:init(self.skillBar.frame, L("Skill") .. ":")
    self.skillBar.label.frame:SetPoint("TOPLEFT", self.frame, "TOPLEFT", 15, -18)
    
    self.skillBar.count = private.objects.ui.countlabel()
    self.skillBar.count:init(self.skillBar.frame)
    self.skillBar.count.frame:SetPoint("TOPLEFT", self.skillBar.frame, "TOPRIGHT", 15, -2)
    self.skillBar.count:SetRight(private.max_skill)
    
    self.skillBar.current = private.objects.ui.bar()
    self.skillBar.current:init(self.skillBar.frame, yellow)
    self.skillBar.current.frame:SetPoint("TOPLEFT", self.skillBar.frame, "BOTTOMLEFT")
    self.skillBar.current:SetSize(300, 4)
    
    -- Set up the skill bar's initial fill and self-update
    self.skillBar:RegisterEvent("CHAT_MSG_SKILL")
    function self.skillBar:CHAT_MSG_SKILL()
        -- Find the maximum possible number of skill points based on the current expansion
        local expMax = private.max_skill
        local _, _, index = GetProfessions()
        if not index then return end -- The character does not have archaeology
        local _, _, skill, max = GetProfessionInfo(index)
        
        if skill == expMax then
            -- Skill can't be increased further! Hide the bar.
            self.frame:Hide()
        else
            -- Set the main bar
            self:SetFill(skill / expMax)
            self.count:SetBoth(skill, expMax)
            
            -- Set the current skill bar
            if max == 75 then
                self.current:SetFill((skill - max + 75) / 75)
            else
                self.current:SetFill((skill - max + 100) / 100)
            end
            if max - skill <= 15 then
                self.current:SetColour(red)
                -- TODO: Warn the player about the skill getting close to the cap
            else
                self.current:SetColour(yellow)
            end
        end
    end
    -- Call the update function once to set the bar's initial length
    self.skillBar:CHAT_MSG_SKILL()
    
    self.raceLabel = private.objects.ui.label()
    self.raceLabel:init(self.frame, "|c00FFFF00" .. L("Race") .. "|r", 16)
    self.raceLabel.frame:SetPoint("TOPLEFT", 15, -37)
    
    self.progressLabel = private.objects.ui.label()
    self.progressLabel:init(self.frame, "|c00FFFF00" .. L("Progress") .. "|r", 16)
    self.progressLabel.frame:SetPoint("TOPLEFT", 115, -37)
    
    self.artifactLabel = private.objects.ui.label()
    self.artifactLabel:init(self.frame, "|c00FFFF00" .. L("Artifact") .. "|r", 16)
    self.artifactLabel.frame:SetPoint("TOPLEFT", 285, -37)
    
    self.keystoneLabel = private.objects.ui.label()
    self.keystoneLabel:init(self.frame, "|c00FFFF00" .. L("Keystones") .. "|r", 16)
    self.keystoneLabel.frame:SetPoint("TOPRIGHT", -15, -37)
    
    -- Travel frame
    self.travelFrame = private.objects.ui.sideFrame()
    self.travelFrame:init(self.frame)
    self.travelFrame.frame:SetPoint("TOPRIGHT", self.frame, "TOPLEFT")
    for k, v in pairs(private.travel.spells) do
        self.travelFrame:addSpell(unpack(v))
    end
    for k, v in pairs(private.travel.items) do
        self.travelFrame:addItem(unpack(v))
    end
    
    local travelFrame = self.travelFrame -- To allow the click function to point to it
    
    -- Travel frame button
    self.travelFrameButton = private.objects.ui.textureButton()
    self.travelFrameButton:init(self.frame, "Interface/Icons/Ability_Mount_FossilizedRaptor", L("Travel frame"), L("Click to open the travel frame"))
    self.travelFrameButton.frame:SetPoint("BOTTOMLEFT", 10, 10)
    function self.travelFrameButton:click()
        travelFrame:Toggle()
    end
    -- Disable the button while in combat
    self.travelFrameButton:RegisterEvent("PLAYER_REGEN_DISABLED")
    self.travelFrameButton:RegisterEvent("PLAYER_REGEN_ENABLED")
    function self.travelFrameButton:PLAYER_REGEN_DISABLED()
        self.frame:Disable()
    end
    function self.travelFrameButton:PLAYER_REGEN_ENABLED()
        self.frame:Enable()
    end
    
    -- Item frame
    self.itemFrame = private.objects.ui.sideFrame()
    self.itemFrame:init(self.frame)
    self.itemFrame.frame:SetPoint("TOPRIGHT", self.frame, "TOPLEFT")
    for k, v in pairs(private.onUseItems) do
        self.itemFrame:addItem(unpack(v))
    end
    
    local itemFrame = self.itemFrame -- To allow the click function to point to it
    
    -- Item frame button
    self.itemFrameButton = private.objects.ui.textureButton()
    self.itemFrameButton:init(self.frame, "Interface/Icons/inv_misc_archaeology_vrykuldrinkinghorn", L("Item frame"), L("Click to open the item frame"))
    self.itemFrameButton.frame:SetPoint("BOTTOMLEFT", 46, 10)
    function self.itemFrameButton:click()
        itemFrame:Toggle()
    end
    -- Disable the button while in combat
    self.itemFrameButton:RegisterEvent("PLAYER_REGEN_DISABLED")
    self.itemFrameButton:RegisterEvent("PLAYER_REGEN_ENABLED")
    function self.itemFrameButton:PLAYER_REGEN_DISABLED()
        self.frame:Disable()
    end
    function self.itemFrameButton:PLAYER_REGEN_ENABLED()
        self.frame:Enable()
    end
    
    -- Crateables frame
    self.crateablesFrame = private.objects.ui.sideFrame()
    self.crateablesFrame:init(self.frame)
    self.crateablesFrame.frame:SetPoint("TOPRIGHT", self.frame, "TOPLEFT")
    for k, v in pairs(private.crateables) do
        self.crateablesFrame:addItem(unpack(v))
    end
    
    local crateablesFrame = self.crateablesFrame -- To allow the click function to point to it
    
    -- Crateables frame button
    self.crateablesFrameButton = private.objects.ui.textureButton()
    self.crateablesFrameButton:init(self.frame, "Interface/Icons/inv_crate_04", L("Crateables frame"), L("Click to open a frame with all your crateable artifacts"))
    self.crateablesFrameButton.frame:SetPoint("BOTTOMLEFT", 82, 10)
    function self.crateablesFrameButton:click()
        crateablesFrame:Toggle()
    end
    -- Disable the button while in combat
    self.crateablesFrameButton:RegisterEvent("PLAYER_REGEN_DISABLED")
    self.crateablesFrameButton:RegisterEvent("PLAYER_REGEN_ENABLED")
    function self.crateablesFrameButton:PLAYER_REGEN_DISABLED()
        self.frame:Disable()
    end
    function self.crateablesFrameButton:PLAYER_REGEN_ENABLED()
        self.frame:Enable()
    end
    -- Close and re-open when a spell finishes casting (to update what items are shown)
    self.crateablesFrameButton:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
    function self.crateablesFrameButton:UNIT_SPELLCAST_SUCCEEDED()
    end
    
    -- Crate counter
    self.crateCounter = private.objects.ui.label()
    self.crateCounter:init(self.frame, ": " .. GetItemCount(87399), 12)
    self.crateCounter.frame:SetPoint("BOTTOMLEFT", 118, 16)
    self.crateCounter:RegisterEvent("BAG_UPDATE")
    function self.crateCounter:BAG_UPDATE()
        self:SetText(": " .. GetItemCount(87399))
        crateablesFrame:Toggle()
        crateablesFrame:Toggle()
    end
    
    -- Survey button
    self.surveyButton = private.objects.ui.spellButton()
    self.surveyButton:init(self.frame, 80451, 3)
    self.surveyButton.frame:SetPoint("BOTTOM", 0, 10)
    
    -- Artifact list frame
    self.artifactList = private.objects.ui.artifactList()
    self.artifactList:init(self.frame)
    
    -- Minimap tracker options button/frame
    self.shovelButton = CreateFrame("Button", nil, self.frame)
    self.shovelButton:SetSize(16, 16)
    self.shovelButton:SetNormalTexture("Interface/Archeology/Arch-Icon-Marker")
    self.shovelButton:SetPoint("BOTTOMRIGHT", -90, 20)
    self.shovelButton.__parent = self
    self.shovelButton:SetScript("OnClick", function(self)
        if self.__parent.trackerOptions:IsShown() then
            self.__parent.trackerOptions:Hide()
        else
            self.__parent.trackerOptions:Show()
        end
    end)
    self.trackerOptions = CreateFrame("Frame", nil, self.shovelButton)
    self.trackerOptions:SetBackdrop(frameBackdrop)
    self.trackerOptions:SetSize(100, 64)
    self.trackerOptions:SetPoint("BOTTOMRIGHT", self.shovelButton, "BOTTOMLEFT")
    self.trackerOptions:SetFrameLevel(5)
    self.trackerOptions:EnableMouse(true)
    self.trackerOptions:Hide()
    self.trackerOptions.track = CreateFrame("CheckButton", "Excavatinator_TrackerCheckButton1", self.trackerOptions, "OptionsCheckButtonTemplate")
    self.trackerOptions.track:SetPoint("BOTTOMLEFT", 8, 8)
    self.trackerOptions.track:SetHitRectInsets(0, -60, 0, 0)
    Excavatinator_TrackerCheckButton1Text:SetText(L("Track"))
    self.trackerOptions.show = CreateFrame("CheckButton", "Excavatinator_TrackerCheckButton2", self.trackerOptions, "OptionsCheckButtonTemplate")
    self.trackerOptions.show:SetHitRectInsets(0, -60, 0, 0)
    self.trackerOptions.show:SetPoint("BOTTOMLEFT", 8, 28)
    Excavatinator_TrackerCheckButton2Text:SetText(L("Show"))
    local function onClick_track(self)
        ExcavatinatorNodeDB.track = self:GetChecked()
    end
    local function onClick_show(self)
        ExcavatinatorNodeDB.show = self:GetChecked()
    end
    self.trackerOptions.track:SetScript("OnClick", onClick_track)
    self.trackerOptions.track:SetChecked(ExcavatinatorNodeDB.track)
    self.trackerOptions.show:SetScript("OnClick", onClick_show)
    self.trackerOptions.show:SetChecked(ExcavatinatorNodeDB.show)
    
    -- Set up the race bars
    local function raceClick(raceBar)
        -- Left click on a race - open the artifact list
        self.artifactList:PopulateRace(raceBar)
    end
    self.races = {}
    for i = 1, #private.artifacts do
        self.races[i] = private.objects.ui.raceBar()
        self.races[i]:init(i, self.frame)
        self.races[i].leftclick = raceClick
        self.races[i]:update()
    end
    
    -- Achievement frame
    self.achievementFrame = private.objects.ui.achievements()
    self.achievementFrame:init(self.frame)
    
    local achievementFrame = self.achievementFrame
    
    -- Links frame
    self.linksFrame = private.objects.ui.linksFrame()
    self.linksFrame:init(self.frame)
    
    local linksFrame = self.linksFrame
    
    -- Achievement button
    self.achieveFrameButton = private.objects.ui.textureButton()
    self.achieveFrameButton:init(self.frame, "Interface/Icons/trade_archaeology_highborne_artifactfragment", L("Achievement frame"), L("Click to open the achievement frame"))
    self.achieveFrameButton.frame:SetPoint("BOTTOMRIGHT", -10, 10)
    function self.achieveFrameButton:click()
        linksFrame:Hide()
        achievementFrame:Toggle()
    end
    
    -- Links frame button
    self.linksFrameButton = private.objects.ui.textureButton()
    self.linksFrameButton:init(self.frame, "Interface/Icons/INV_Belt_18", L("Links frame"), L("Click to open the links frame"))
    self.linksFrameButton.frame:SetPoint("BOTTOMRIGHT", -46, 10)
    function self.linksFrameButton:click()
        achievementFrame:Hide()
        linksFrame:Toggle()
    end
    
    -- Set up the bottom bar
    self.bottom = {}
    self.bottom.bar = private.objects.ui.bar()
    self.bottom.bar:init(self.frame)
    self.bottom.bar.frame:SetPoint("BOTTOMLEFT", 120, 55)
    self.bottom.bar:SetSize(350, 10)
    
    self.bottom.label = private.objects.ui.label()
    self.bottom.label:init(self.frame, L("Total progress") .. ":")
    self.bottom.label.frame:SetPoint("BOTTOMLEFT", 15, 55)
    
    self.bottom.button = CreateFrame("Button", nil, self.frame)
    self.bottom.button.__parent = self
    self.bottom.button:SetSize(455, 10)
    self.bottom.button:SetPoint("BOTTOMLEFT", 15, 55)
    self.bottom.button:RegisterForClicks("LeftButtonUp", "RightButtonUp")
    self.bottom.button:SetScript("OnClick", function(self, button)
        if button == "LeftButton" then
            self.__parent.artifactList:PopulateMissing(self.__parent.races, true, true)
        end
        if button == "RightButton" then
            self.__parent.artifactList:PopulateAll(self.__parent.races, true, true)
        end
    end)
    self.bottom.button:SetScript("OnEnter", function(self)
        GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
        GameTooltip:ClearLines()
        GameTooltip:AddLine(format(L("%d / %d artifacts completed."), self.__parent.bottom.collected, self.__parent.bottom.total))
        GameTooltip:AddLine(L("Click to see a list of incomplete artifacts."), 0, 1, 0)
        GameTooltip:AddLine(L("Right click to see all artifacts."), 0, 1, 0)
        GameTooltip:Show()
    end)
    self.bottom.button:SetScript("OnLeave", function(self)
        GameTooltip:Hide()
    end)
    
    self.bottom.collected = 0
    self.bottom.total = 0
    
    self:UpdateBar()
    
    -- Set up the expand/collapse function
    self.expand = CreateFrame("Button", nil, self.frame)
    self.expand.__parent = self
    self.expand:SetNormalTexture("Interface/Buttons/UI-Panel-ExpandButton-Up")
    self.expand:SetPushedTexture("Interface/Buttons/UI-Panel-ExpandButton-Down")
    self.expand:SetHighlightTexture("Interface/Buttons/UI-Panel-MinimizeButton-Highlight")
    self.expand:SetSize(32, 32)
    self.expand:SetPoint("TOPRIGHT", -37, -5)
    self.expand:SetScript("OnClick", function(self)
        self.__parent:ShowAll()
    end)
    self.expand:Hide()
    self.collapse = CreateFrame("Button", nil, self.frame)
    self.collapse.__parent = self
    self.collapse:SetNormalTexture("Interface/Buttons/UI-Panel-CollapseButton-Up")
    self.collapse:SetPushedTexture("Interface/Buttons/UI-Panel-CollapseButton-Down")
    self.collapse:SetHighlightTexture("Interface/Buttons/UI-Panel-MinimizeButton-Highlight")
    self.collapse:SetSize(32, 32)
    self.collapse:SetPoint("TOPRIGHT", -37, -5)
    self.collapse:SetScript("OnClick", function(self)
        self.__parent:ShowContinent()
    end)
    
    -- Set up the close button
    self.closeButton = CreateFrame("Button", nil, self.frame, "UIPanelCloseButton")
    self.closeButton:SetPoint("TOPRIGHT", -5, -5)
    
    -- Register to listen for fragment finds
    self:RegisterEvent("CHAT_MSG_CURRENCY")
    self:RegisterEvent("BAG_UPDATE")
    self:RegisterEvent("ARTIFACT_COMPLETE")
    self:RegisterEvent("ARTIFACT_HISTORY_READY")
    self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
    self.frame:SetScript("OnUpdate", self.OnUpdate)
    
    self.isCollapsed = false
    self:ShowAll()
end

function private.objects.ui.raceFrame:ShowContinent(continent)
    -- Changed to avoid a lot of logic regarding race 10, "Other".
    -- What was previously the failsafe data for continent 0, is now the
    -- list of races to display when expanded. It will still act as a
    -- failsafe when the list is collapsed to only the current continent.
    if not continent then
        -- Retrieve the current continent
        SetMapToCurrentZone()
        continent = GetCurrentMapContinent()
    end
    
    self.isCollapsed = true
    if private.continents[continent] then
        -- Hide them all first
        for i = 1, #self.races do
            self.races[i].frame:Hide()
        end
        for i = 1, #private.continents[continent] do
            self.races[private.continents[continent][i]].frame:Show()
            self.races[private.continents[continent][i]].frame:SetPoint("TOP", 0, -40 - (20 * i))
        end
        self.frame:SetHeight(135 + 20 * #private.continents[continent])
    else
        self:ShowAll()
    end
    self.collapse:Hide()
    self.expand:Show()
end

function private.objects.ui.raceFrame:ShowAll()
    -- Changed to avoid a lot of logic regarding race 10, "Other".
    -- What was previously the failsafe data for continent 0, is now the
    -- list of races to display when expanded. It will still act as a
    -- failsafe when the list is collapsed to only the current continent.
    self:ShowContinent(0)
    self.isCollapsed = false
    self.expand:Hide()
    self.collapse:Show()
end

function private.objects.ui.raceFrame:UpdateBar()
    -- Updates the total progress bar at the bottom of the frame
    self.bottom.collected, self.bottom.total = 0, 0
    for k, v in pairs(self.races) do
        self.bottom.collected = self.bottom.collected + v.raceObject.completed.total
        self.bottom.total = self.bottom.total + v.raceObject.numArtifacts.total
    end
    
    self.bottom.bar:SetFill(self.bottom.collected / self.bottom.total)
end

function private.objects.ui.raceFrame:Show()
    -- If in combat, print an error and do nothing.
    if InCombatLockdown() then
        print(L("You are in combat! Excavatinator cannot open."))
        return nil
    end
    -- Updates the bar,
    self:UpdateBar()
    -- then shows the frame
    self.frame:Show()
end

function private.objects.ui.raceFrame:Hide(...)
    return self.frame:Hide(...)
end

function private.objects.ui.raceFrame:IsShown(...)
    return self.frame:IsShown(...)
end

-- Events for the race frame

local allow_bag_update = true
function private.objects.ui.raceFrame:OnUpdate()
    -- This is a workaround. The bag update event fires hundreds of times in certain situations, yet I rely on it.
    -- I'm therefore only allowing it to run once every update.
    allow_bag_update = true
end

function private.objects.ui.raceFrame:CHAT_MSG_CURRENCY(currencyString)
    -- This fires every time a currency is added.
    -- The currency ID is checked against the known fragment IDs,
    -- and the appropriate race is updated.
    local currencyID = currencyString:match("Hcurrency:(%d+)")
    local raceID = private.currencies[tonumber(currencyID)]
    if raceID then self.races[raceID]:update() end
    self:UpdateBar()
end

function private.objects.ui.raceFrame:ARTIFACT_COMPLETE()
    -- This should really not be necessary.
    -- Every time an artifact is completed, the completion count on
    -- that artifact is set to 0, even if you're viewing it by artifact
    -- ID and not just looking at the current artifact.
    -- The only known way to fix this is to re-request the artifact
    -- completion history every time. I can't imagine the servers
    -- love that, but it seems to be by design.
    RequestArtifactCompletionHistory()
end

function private.objects.ui.raceFrame:ARTIFACT_HISTORY_READY()
    for k, v in pairs(self.races) do
        v:update()
    end
    self:UpdateBar()
end

function private.objects.ui.raceFrame:BAG_UPDATE()
    if not allow_bag_update then return end
    allow_bag_update = false
    -- Triggers an update whenever an item is received or lost/used.
    -- Apparently, ARTIFACT_COMPLETE and even ARTIFACT_HISTORY_READY
    -- both trigger far too early to get the name of the next project.
    -- This also has the added benefit of updating the display every
    -- time a keystone is picked up.
    for k, v in pairs(self.races) do
        v:update()
    end
    self:UpdateBar()
end

function private.objects.ui.raceFrame:ZONE_CHANGED_NEW_AREA()
    if self.isCollapsed then
        self:ShowContinent()
    end
end